#include "MeshLoader.h"
#include "CustomHash.h"

const std::vector<Node>& MeshLoader::Get_node() const
{
	return nodes;
}

const std::vector<Finite_Element>& MeshLoader::Get_finite_elems() const
{
	return finite_elems;
}

const std::vector<Boundary_Finite_Element>& MeshLoader::Get_boundary() const
{
	return boundary;
}

std::vector<int> MeshLoader::Get_elems_by_ID(int p_node_id1, int p_node_id2, int p_node_id3) const
{
	std::vector<int> res;
	auto first_to_find = finite_elems.begin();
	auto last_to_find = finite_elems.end();
	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[p_node_id1, p_node_id2, p_node_id3](const Finite_Element& elem) {
				return std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id1) != elem.node_id.end()
					&& std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id2) != elem.node_id.end()
					&& std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id3) != elem.node_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}
	return res;
}

std::vector<int> MeshLoader::Get_elems_by_edge(int p_node_id1, int p_node_id2) const
{
	std::vector<int> res;
	auto first_to_find = finite_elems.begin();
	auto last_to_find = finite_elems.end();
	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[p_node_id1, p_node_id2](const Finite_Element& elem) {
				return std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id1) != elem.node_id.end()
					&& std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id2) != elem.node_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}
	return res;
}

std::vector<int> MeshLoader::Get_elems_by_material(int p_material_id) const
{
	std::vector<int> res;
	auto first_to_find = finite_elems.begin();
	auto last_to_find = finite_elems.end();
	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[p_material_id](const Finite_Element& elem) {
				return elem.material_id == p_material_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}
	return res;
}

std::set<int> MeshLoader::Get_nodes_by_surface(int p_surface_id) const
{
	std::set<int> res;
	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();
	while (first_to_find != last_to_find) {
		auto found_boundary = std::find_if(first_to_find, last_to_find,
			[p_surface_id](const Boundary_Finite_Element& cur_boundary) {
				return cur_boundary.surface_id == p_surface_id;
			});

		first_to_find = found_boundary;
		if (found_boundary != last_to_find) {
			for (const auto& node_in_boundary : found_boundary->node_id)
				res.insert(node_in_boundary);
			++first_to_find;
		}
	}
	return res;
}

std::vector<int> MeshLoader::Get_boundary_by_surface(int p_surface_id) const
{
	std::vector<int> res;
	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();
	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[p_surface_id](const Boundary_Finite_Element& elem) {
				return elem.surface_id == p_surface_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}
	return res;
}

void MeshLoader::Insert_new_nodes()
{
	std::unordered_set <Edge> examined_edges;
	for (auto& elem : finite_elems) {
		std::vector<int> cur_nodes_id = elem.node_id;
		for (auto first_node = 0; first_node < 4; ++first_node)
			for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
				Edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);

				if (examined_edges.insert(cur_edge).second) { 
					Node new_node = Get_middle_node(cur_edge);
					cur_edge.Update_Middle_Node(new_node.node_id);
					nodes.push_back(new_node);
				}
				else {
					elem.node_id.push_back(cur_edge.middle_node);
				}
			}
	}

	for (auto& elem : boundary) {
		std::vector<int> cur_nodes_id = elem.node_id;
		for (auto first_node = 0; first_node < 3; ++first_node)
			for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
				auto found_edge = examined_edges.find({ cur_nodes_id[first_node], cur_nodes_id[second_node] });
				if (found_edge == examined_edges.end())
					found_edge = examined_edges.find({ cur_nodes_id[second_node], cur_nodes_id[first_node] });
				elem.node_id.push_back(found_edge->middle_node);
			}
	}
}

std::vector<std::set<int>> MeshLoader::Get_adjacent_nodes() const
{
	std::vector<std::set<int>> res(nodes.size() + 1);
	for (const auto& cur_elem : finite_elems) {
		for (auto cur_node_id : cur_elem.node_id)
			for (auto adj_node_id : cur_elem.node_id)
				if (cur_node_id != adj_node_id)
					res[cur_node_id].insert(adj_node_id);
	}
	return res;
}

std::array<double, 3> MeshLoader::Get_middle_of_edge(const Edge& p_edge)
{
	return
	{ (get<0>(nodes[p_edge.nodes.first - 1].coors) + get<0>(nodes[p_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<1>(nodes[p_edge.nodes.first - 1].coors) + get<1>(nodes[p_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<2>(nodes[p_edge.nodes.first - 1].coors) + get<2>(nodes[p_edge.nodes.second - 1].coors)) * 1 / 2 };
}

Node MeshLoader::Get_middle_node(const Edge& cur_edge)
{
	Node new_node(nodes.size() + 1, Get_middle_of_edge(cur_edge), 0);
	return new_node;
}

Node& MeshLoader::Get_node_by_ID(int p_id)
{
	return nodes.at(p_id);
}

std::array<double, 3> MeshLoader::Get_coors_node_by_ID(int p_id)
{
	return nodes.at(p_id).coors;
}












